
/* Ryan Drake (stiletto@mediaone.net)
 *
 * To build:  gcc -Wall -o png2empeg png2empeg.c -lpng
 *
 * To run:
 *   Here is a bash script that will take one or two PNG files
 *   and load them to the empeg player.  USE AT YOUR OWN RISK!!
 *
 * HOME=$1
 * if [ -n $2 ] ; then AUTO=$1 ; else AUTO=$2 ; fi
 * echo -n empg > .tmpempeg
 * ./png2empeg < $HOME >> .tmpempeg
 * ./png2empeg < $AUTO >> .tmpempeg
 * ./download .tmpempeg a000
 * rm -f .tmpempeg
 *
 */

#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <png.h>

#define _GNU_SOURCE
#include <getopt.h>

static int cutoff0 = 30;
static int cutoff1 = 93;
static int cutoff2 = 200;
static char vermsg[] = "png2empeg version 0.1\n";
static char helpmsg[] = "Usage: png2empeg [OPTION]...\n"
                        "Convert grayscale PNG file into a format suitable\n"
                        "for display on an \"empeg\" unit.\n\n"
                        "  -0, --cutoff0=N  cutoff between black and gray0\n"
                        "  -1, --cutoff1=N  cutoff between gray0 and gray1\n"
                        "  -2, --cutoff2=N  cutoff between gray1 and white\n"
                        "  -h, --help       this help message\n"
                        "  -V, --version    output version info and exit\n";

static struct option long_options[] = {
    { "version", 0, 0, 'V' },
    { "help", 0, 0, 'h' },
    { "cutoff0", 1, 0, '0' },
    { "cutoff1", 1, 0, '1' },
    { "cutoff2", 1, 0, '2' },
    { 0, 0, 0, 0 }
};

#define PNG_BYTES_TO_CHECK 8
int
main(int argc, char** argv)
{
    png_structp png_ptr;
    png_infop info_ptr;
    png_uint_32 w, h, rowbytes;
    int bit_depth, color_type, row;
    char buf[PNG_BYTES_TO_CHECK];
    png_bytep* row_pointers;

    unsigned char* d;
    unsigned short* sd;
    char outc;
    int i, ind;

    /* Read options from the command line */
    while((i = getopt_long(argc, argv, "Vh0:1:2:", long_options, &ind)) != -1) {
        switch(i) {
            case '0': cutoff0 = strtol(optarg, NULL, 10); break;
            case '1': cutoff1 = strtol(optarg, NULL, 10); break;
            case '2': cutoff2 = strtol(optarg, NULL, 10); break;
            case 'V': fputs(vermsg, stderr); exit(EXIT_SUCCESS);
            case 'h':
            default:  fputs(helpmsg, stderr); exit(EXIT_SUCCESS);
        }
    }

    /* Read in some of the signature bytes */
    if (fread(buf, 1, PNG_BYTES_TO_CHECK, stdin) != PNG_BYTES_TO_CHECK) {
        fputs("cannot read PNG signature\n", stderr);
        exit(EXIT_FAILURE);
    }

    /* Compare the first PNG_BYTES_TO_CHECK bytes of the signature */
    if (png_sig_cmp(buf, (png_size_t)0, PNG_BYTES_TO_CHECK)) {
        fputs("incorrect PNG signature\n", stderr);
        exit(EXIT_FAILURE);
    }

    /* Create "read struct" */
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (png_ptr == NULL) {
        fclose(stdin);
        fputs("out of memory\n", stderr);
        exit(EXIT_FAILURE);
    }

    /* Create "info struct" */
    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL) {
        png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
        fclose(stdin);
        fputs("out of memory\n", stderr);
        exit(EXIT_FAILURE);
    }

    /* Set error handling */
    if (setjmp(png_ptr->jmpbuf)) {
        png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
        fclose(stdin);
        fputs("cannot setjmp\n", stderr);
        exit(EXIT_FAILURE);
    }

    /* Set up the input control */
    png_init_io(png_ptr, stdin);

    /* If we have already read some of the signature */
    png_set_sig_bytes(png_ptr, PNG_BYTES_TO_CHECK);

    /* Get info about PNG file */
    png_read_info(png_ptr, info_ptr);
    png_get_IHDR(png_ptr, info_ptr, &w, &h, &bit_depth, &color_type,
         NULL, NULL, NULL);
    if((w != 128) || (h != 32)) {
        png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
        fclose(stdin);
        fputs("dimentions of PNG are not 128x32\n", stderr);
        exit(EXIT_FAILURE);
    }

    /* Expand paletted colors into true RGB triplets */
    if (color_type == PNG_COLOR_TYPE_PALETTE) {
        png_set_expand(png_ptr);
    }

    /* Expand grayscale images to the full 8 bits from 1, 2, or 4 bits/pixel */
    if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8) {
        png_set_expand(png_ptr);
    }

    /* Expand paletted or RGB images with transparency to full alpha channels
     * so the data will be available as RGBA quartets.
     */
    if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS)) {
        png_set_expand(png_ptr);
    }

    /* Allocate the memory to hold the image using the fields of info_ptr. */

    row_pointers = calloc (h, sizeof(*row_pointers));
    if (row_pointers == NULL) {
        png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
        fclose(stdin);
        fputs("out of memory\n", stderr);
        exit(EXIT_FAILURE);
    }

    rowbytes = png_get_rowbytes (png_ptr, info_ptr);
    row_pointers[0] = calloc (h, rowbytes);
    if (row_pointers[0] == NULL) {
        free (row_pointers);
        png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
        fclose(stdin);
        fputs("out of memory\n", stderr);
        exit(EXIT_FAILURE);
    }

    for (row = 1; row < h; row++) {
        row_pointers[row] = row_pointers[row-1] + rowbytes;
    }

    png_read_image(png_ptr, row_pointers);
    png_read_end(png_ptr, info_ptr);

    /* Set pointer so we can find the data later */
    d = (char*)row_pointers[0];

    /* Clean up */
    free(row_pointers);
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    fclose(stdin);

    /* Select format and components */
    switch(color_type) {
        case PNG_COLOR_TYPE_GRAY:

            switch(bit_depth) {
                case 8:
                    /* OK */
                    break;
                case 16:
                    /* chop to 8 bpp */
                    sd = (unsigned short*)d;
                    for(i=0; i<w*h; i++) { d[i] = sd[i] >> 8; }
                    break;
                default:
                    /* Unsupported */
                    free(d);
                    fputs("unsupported grayscale PNG format\n", stderr);
                    exit(EXIT_FAILURE);
                    break;
            }
            break;

        case PNG_COLOR_TYPE_GRAY_ALPHA:
        case PNG_COLOR_TYPE_RGB:
        case PNG_COLOR_TYPE_RGB_ALPHA:
        default:
            /* Unsupported */
            free(d);
            fputs("unsupported PNG format\n", stderr);
            exit(EXIT_FAILURE);
            break;
    }

    /* Write output file 2048 bytes */
    for(i=0; i<w*h; i++) {

        if(d[i] < cutoff0) outc = 0;
        else if(d[i] < cutoff1) outc = 1;
        else if(d[i] < cutoff2) outc = 2;
        else outc = 3;

        i++;

        if(d[i] < cutoff0) outc |= 0;
        else if(d[i] < cutoff1) outc |= 16;
        else if(d[i] < cutoff2) outc |= 32;
        else outc |= 64;

        putchar(outc);
    }

    free(d);
    exit(EXIT_SUCCESS);
}
